/*
 * stack.c
 *
 *  Created on: 18 Oct 2011
 *  Copyright 2011 Nicolas Melot
 *
 * This file is part of TDDD56.
 *
 *     TDDD56 is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     TDDD56 is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with TDDD56. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef DEBUG
#define NDEBUG
#endif

#include <assert.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>

#include "stack.h"
#include "non_blocking.h"

#if NON_BLOCKING == 0
#warning Stacks are synchronized through locks
#else
#if NON_BLOCKING == 1
#warning Stacks are synchronized through hardware CAS
#else
#warning Stacks are synchronized through lock-based CAS
#endif
#endif
pthread_mutex_t lock;
int
stack_check(stack_t *stack, int gt)
{
// Do not perform any sanity check if performance is bein measured
#if MEASURE == 0
	// Use assert() to check if your stack is in a state that makes sens
	int size = stack_size(stack);
	/*
	#if MEASURE == 1
	int gt = MAX_PUSH_POP-((MAX_PUSH_POP/NB_THREADS)*NB_THREADS);
	#elif MEASURE == 2
	int gt = ((MAX_PUSH_POP/NB_THREADS)*NB_THREADS);
	#endif
	*/
	if(size == gt) {
		printf("%s\n", "Stack is in good state! :)");
		printf("%s %i %s %i\n", "Counter:", size, "Size:", gt);
	}
	else {
		printf("%s%i%s%i\n", "Stack is in bad state! :( No of nodes is ", size, " but should be ", gt);
	}

	// This test fails if the task is not allocated or if the allocation failed
	assert(stack != NULL);
#endif
	// The stack is always fine
	return 1;
}

int stack_size(stack_t* stack) {
	node_t* tmp = stack->head;

  int counter = 0;
  while(tmp) {
    tmp = tmp->next;
    counter++;
  }
	return counter;
}

void stack_init(stack_t* stack, int no_of_nodes) {
	pthread_mutex_init(&lock, NULL);
	stack->head = NULL;
	stack->size = 0;

	for(int i = 0; i < no_of_nodes; i++)
	{
		allocate_node(stack);
	}
}

node_t* stack_pop_free(stack_t* stack) {
	if(stack->head == NULL) {
		return allocate_node(stack);
	}

	node_t* tmp = stack->head;
	stack->head = stack->head->next;
	stack->size--;

	return tmp;
}

void stack_push_free(stack_t* stack, node_t* node) {
	node->next = stack->head;
	stack->head = node;
	stack->size++;
}

void /* Return the type you prefer */
stack_push(stack_t *stack, stack_t *free_stack, int data)
{
	node_t* newNode = stack_pop_free(free_stack);
	newNode->value = data;
#if NON_BLOCKING == 0
  // Implement a lock_based stack
	pthread_mutex_lock(&lock);
	newNode->next = stack->head;
	stack->head = newNode;
	stack->size++;
	pthread_mutex_unlock(&lock);

#elif NON_BLOCKING == 1
  // Implement a harware CAS-based stack
	node_t* head;
	do {
	 head = stack->head;
	 newNode->next = head;
 } while(cas((size_t*)&stack->head, (size_t)head, (size_t)newNode) != (size_t)head);

#else
  /*** Optional ***/
  // Implement a software CAS-based stack
#endif

  // Debug practice: you can check if this operation results in a stack in a consistent check
  // It doesn't harm performance as sanity check are disabled at measurement time
  // This is to be updated as your implementation progresses

  //stack_check(stack);

}
int /* Return the type you prefer */
stack_pop(stack_t* stack, stack_t* free_stack)
{
#if NON_BLOCKING == 0
  // Implement a lock_based stack

	if(stack->head == NULL) {
		return -1;
	}
	pthread_mutex_lock(&lock);
	node_t* tmp = stack->head;
	stack->head = stack->head->next;
	stack->size--;
	pthread_mutex_unlock(&lock);

	stack_push_free(free_stack, tmp);

	return tmp->value;

#elif NON_BLOCKING == 1
  // Implement a harware CAS-based stack

	node_t *head;
	node_t *headnext;
	do {
		if (stack->head == NULL) {
			return NULL;
		}
		head = stack->head;
		headnext = stack->head;
	} while (cas((size_t*)&stack->head, (size_t)head, (size_t)headnext->next) != (size_t)head);
	stack_push_free(free_stack, head);

	//stack_check(stack);
	return head->value;

#else
  /*** Optional ***/
  // Implement a software CAS-based stack
#endif

  return 0;
}

void free_nodes(stack_t* stack)
{
	while(stack->head != NULL && stack->head->next != NULL) {
  	node_t* tmp = stack->head;
  	stack->head = stack->head->next;
		stack->size--;
  	free(tmp);
  }
}

void print_sizes(stack_t* stack)
{
	pthread_mutex_lock(&lock);
	printf("%i\n", stack->size);
	pthread_mutex_unlock(&lock);
}
node_t* allocate_node(stack_t* stack){
	node_t* newNode =  malloc(sizeof(node_t));
	newNode->next = stack->head;
	stack->head = newNode;
	stack->size++;
	return newNode;
}
#if NON_BLOCKING == 1
void stack_push_ABA(stack_t *stack, stack_t *free_stack, int id, int* flag)
{
		while(1) {
			//printf("%s%i%s\n","Thread ", id, " is waiting..." );
			//printf("%s%i\n", "Flag is ", *flag);
			if(*flag == 4 && 1 == id)
				break;
		}
		printf("%s%i\n", "Pushing with thread ", id);
		node_t* newNode = stack_pop_free(free_stack);
		node_t* oldNode;
		do {
		 oldNode = stack->head;
		 newNode->next = oldNode;

	 } while(cas((size_t*)&stack->head, (size_t)oldNode, (size_t)newNode) != (size_t)oldNode);
		stack->size++;
		printf("%s%i\n", "Cas push successful! Value ", newNode->value);
		*flag = 0;

}

int stack_pop_ABA(stack_t *stack, stack_t *free_stack, int id, int* flag)
{
		while(*flag != id) {
			//printf("%s%i%s\n","Thread ", id, " is waiting..." );
			//printf("%s%i\n", "Flag is ", *flag);
		}
		printf("%s%i\n", "Pop call on thread ", id);
		node_t *head;
		node_t *headnext;
		do {
	    head = stack->head;
	    if (stack->head == NULL) {
				printf("%s\n", "Stack head is null!");
	      return NULL;
			}
	    headnext = stack->head->next;
			if(id == 0) {
				printf("%s%i\n", "Thread waiting... id: ", id);
				*flag = 1;
				while(*flag != 0) {
				//	printf("%s\n", "Thread 0 is waiting...");
				}
				printf("%s%i\n", "Thread continues... id: ", id);
			}
			printf("%s%i\n", "Attempting pop CAS on thread ", id);
		} while (cas((size_t*)&stack->head, (size_t)head, (size_t)headnext) != (size_t)head);

			printf("%s%i%s%i\n", "Pop CAS successful! Node value from thread ", id, " is ", head->value);
			stack_push_free(free_stack, head);
			stack->size--;

			if(id == 2) {
				*flag = 4;
			}
			if(id == 1) {
				*flag = 2;
			}
			if(id == 0) {
				*flag = 1;
			}
			return head->value;
}
#endif
