// Mandelbrot explorer, based on my old Julia demo plus parts of Nicolas Melot's Lab 1 code.
// CPU only! Your task: Rewrite for CUDA! Test and evaluate performance.

// Emergency version for use without graphics output. I recommend the interactive version if you can.
// Preliminary! I will add timing using milli.c.

// Compile with:
// g++ noninteractiveMandelbrot.cpp readppm.cpp -o noninteractiveMandelbrot

// Your CUDA version should compile with something like
// nvcc -lglut -lGL interactiveMandelbrotCUDA.cu -o interactiveMandelbrotCUDA
// See lecture notes on how to compile and link with the other code.
// Note: A Simple trick is to #include the other files.

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "readppm.h"
// uses framework Cocoa

// Image data
	unsigned char	*pixels = NULL;
	unsigned char	*pixelsCUDA = NULL;

	#define DIM 512

	int size = DIM * DIM * 4;

	// Select precision here! float or double!
	#define MYFLOAT double

	// User controlled parameters
	int maxiter = 20;
	MYFLOAT offsetx = -200, offsety = 0, zoom = 0;
	MYFLOAT scale = 1.5;

	const int blocksize = 32;
	const int gridsize = 16;

// Init image data
void initBitmap(int width, int height)
{
	pixels = (unsigned char *)malloc(width*height*4);
	for(int i = 0; i < size; i++)
		pixels[i] = 255;
	cudaMalloc( (void**)&pixelsCUDA, size );
	cudaMemcpy( pixelsCUDA, pixels, size, cudaMemcpyHostToDevice );
}

// Complex number class

struct cuComplex
{
    MYFLOAT   r;
    MYFLOAT   i;
		__device__
    cuComplex( MYFLOAT a, MYFLOAT b ) : r(a), i(b)  {}
		__device__
    float magnitude2( void )
    {
        return r * r + i * i;
    }
		__device__
    cuComplex operator*(const cuComplex& a)
    {
        return cuComplex(r*a.r - i*a.i, i*a.r + r*a.i);
    }
		__device__
    cuComplex operator+(const cuComplex& a)
    {
        return cuComplex(r+a.r, i+a.i);
    }
};

__device__
int mandelbrot( int x, int y, int gImageWidth, int gImageHeight, int maxiter,
								MYFLOAT offsetx = -200, MYFLOAT offsety = 0, MYFLOAT scale = 1.5)
{
    MYFLOAT jx = scale * (MYFLOAT)(gImageWidth/2 - x + offsetx/scale)/(gImageWidth/2);
    MYFLOAT jy = scale * (MYFLOAT)(gImageHeight/2 - y + offsety/scale)/(gImageWidth/2);

    cuComplex c(jx, jy);
    cuComplex a(jx, jy);

    int i = 0;
    for (i=0; i<maxiter; i++)
    {
        a = a * a + c;
        if (a.magnitude2() > 1000)
            return i;
    }

    return i;
}

__global__
void computeFractal(unsigned char *data, int gImageWidth, int gImageHeight, int maxiter,
										MYFLOAT offsetx = -200, MYFLOAT offsety = 0,  MYFLOAT zoom = 0, MYFLOAT scale = 1.5)
{
    // map from x, y to pixel position
		int x = blockIdx.x * blockDim.x + threadIdx.x;
		int y = blockIdx.y * blockDim.y + threadIdx.y;
		int index = x + y * gImageWidth;



			int fractalValue = mandelbrot(x,y,gImageWidth, gImageHeight, maxiter);

	    // Colorize it
	    int red = 255 * fractalValue/maxiter;
	    if (red > 255) red = 255 - red;
	    	int green = 255 * fractalValue*4/maxiter;
	    if (green > 255) green = 255 - green;
	    	int blue = 255 * fractalValue*20/maxiter;
	    if (blue > 255) blue = 255 - blue;

			data[index*4 + 0] = red;
    	data[index*4 + 1] = green;
    	data[index*4 + 2] = blue;
 	    data[index*4 + 3] = 255;
}


// Main program, inits
int main( int argc, char** argv)
{
	cudaDeviceProp prop;
  cudaGetDeviceProperties(&prop, 0);
  printf("Device Number: %d\n", 0);
  printf("  Device name: %s\n", prop.name);
  printf("  Memory Clock Rate (KHz): %d\n",
         prop.memoryClockRate);
  printf("  Memory Bus Width (bits): %d\n",
         prop.memoryBusWidth);
  printf("  Peak Memory Bandwidth (GB/s): %f\n",
         2.0*prop.memoryClockRate*(prop.memoryBusWidth/8)/1.0e6);
  printf("  Max threads per block: %i\n",
         prop.maxThreadsPerBlock);
	printf("  Max threads dim: %i x %i x %i\n",
         prop.maxThreadsDim[0],prop.maxThreadsDim[1],prop.maxThreadsDim[2] );
  printf("  Max grid size: %i x %i x %i\n",
         prop.maxGridSize[0],prop.maxGridSize[1],prop.maxGridSize[2] );

	float theTime = 0.0;
	cudaEvent_t startEvent, finEvent;

	initBitmap(DIM, DIM);
	dim3 dimBlock( blocksize, blocksize );
	dim3 dimGrid( gridsize, gridsize );

	cudaEventCreate(&startEvent);
	cudaEventCreate(&finEvent);
	cudaEventRecord(startEvent, 0);

	computeFractal<<<dimGrid, dimBlock>>>(pixelsCUDA, DIM, DIM, maxiter);
	cudaThreadSynchronize();
	cudaError_t err = cudaGetLastError();
	 if (err != cudaSuccess)
			 printf("Error: %s\n", cudaGetErrorString(err));

	cudaEventRecord(finEvent, 0);
	cudaEventSynchronize(finEvent);
	cudaEventElapsedTime(&theTime, startEvent, finEvent);
	printf("%s%f\n", "Time elapsed (ms): ", theTime);

	cudaMemcpy( pixels, pixelsCUDA, size, cudaMemcpyDeviceToHost );
	err = cudaGetLastError();
	 if (err != cudaSuccess)
			 printf("Error: %s\n", cudaGetErrorString(err));

	cudaFree( pixelsCUDA );
	err = cudaGetLastError();
	 if (err != cudaSuccess)
			 printf("Error: %s\n", cudaGetErrorString(err));

// Dump to PPM
	writeppm("fractalout.ppm", DIM, DIM, pixels);
	delete[] pixels;
}
